from typing import Dict, List


def reverse_rotation_vertices(normalized_vertices, rotation_angle) -> List[Dict]:
    """Allows to retrieve vertices without rotation."""
    vertices_before_rotate = []
    for vertice in normalized_vertices:
        new_x = vertice["x"]
        new_y = vertice["y"]
        if rotation_angle == 90:
            new_x = vertice["y"]
            new_y = 1 - vertice["x"]
        elif rotation_angle == 180:
            new_x = 1 - vertice["x"]
            new_y = 1 - vertice["y"]
        elif rotation_angle == 270:
            new_x = 1 - vertice["y"]
            new_y = vertice["x"]
        vertices_before_rotate.append({"x": new_x, "y": new_y})
    return vertices_before_rotate



"""Kili Formats
This module provides functions to convert Kili data formats to various other formats"""

from .format.coco import convert_from_kili_to_coco_format
from .format.llm import (
    convert_from_kili_to_llm_rlhf_format,
    convert_from_kili_to_llm_static_or_dynamic_format,
)
from .format.voc import convert_from_kili_to_voc_format
from .format.yolo import convert_from_kili_to_yolo_format
from .kili import clean_json_response, convert_to_pixel_coords, format_json_response

__all__ = [
    "clean_json_response",
    "convert_from_kili_to_coco_format",
    "convert_from_kili_to_llm_rlhf_format",
    "convert_from_kili_to_llm_static_or_dynamic_format",
    "convert_from_kili_to_voc_format",
    "convert_from_kili_to_yolo_format",
    "convert_to_pixel_coords",
    "format_json_response",
]



from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

from kili_formats.exceptions import NotCompatibleOptions
from kili_formats.kili import convert_to_pixel_coords
from src.kili_formats.tool.annotations_to_json_response import (
    AnnotationsToJsonResponseConverter,
)

from .fakes.image import (
    image_asset,
    image_asset_rotated,
    image_project,
    image_project_asset_normalized,
    image_project_asset_unnormalized,
    image_rotated_project_asset_unnormalized,
)
from .fakes.pdf import (
    pdf_asset,
    pdf_asset_rotated,
    pdf_project,
    pdf_project_asset_unnormalized,
)
from .fakes.text import text_asset, text_project, text_project_asset_unnormalized
from .fakes.video import (
    test_cases,
    video_asset,
    video_project,
    video_project_asset_unnormalized,
)


def test_kili_convert_to_pixel_coords_pdf():
    """Test the conversion of coordinates from normalized to pixel values for PDF files."""
    scaled_asset = convert_to_pixel_coords(pdf_asset, pdf_project, normalized_coordinates=False)
    assert scaled_asset == pdf_project_asset_unnormalized


def test_kili_convert_to_pixel_coords_text():
    """Test the conversion of coordinates from normalized to pixel values for PDF files."""
    scaled_asset = convert_to_pixel_coords(text_asset, text_project, normalized_coordinates=False)
    assert scaled_asset == text_project_asset_unnormalized


def test_kili_convert_to_pixel_coords_pdf_rotated_throw_error():
    """Test the conversion of coordinates from normalized to pixel values for PDF files raises error if rotated."""
    with pytest.raises(NotCompatibleOptions, match="PDF labels cannot be rotated"):
        convert_to_pixel_coords(pdf_asset_rotated, pdf_project, normalized_coordinates=False)


def test_kili_convert_to_pixel_coords_image():
    """Test the conversion of coordinates from normalized to pixel values for image files."""
    scaled_asset = convert_to_pixel_coords(image_asset, image_project, normalized_coordinates=False)
    assert scaled_asset == image_project_asset_unnormalized


def test_kili_convert_to_pixel_coords_image_with_normalized_coordinates():
    """Test the conversion of coordinates from normalized to pixel values for image files."""
    scaled_asset = convert_to_pixel_coords(image_asset, image_project, normalized_coordinates=True)
    assert scaled_asset == image_project_asset_normalized


def test_kili_convert_to_pixel_coords_video():
    """Test the conversion of coordinates from normalized to pixel values for video files."""
    scaled_asset = convert_to_pixel_coords(video_asset, video_project, normalized_coordinates=False)
    assert scaled_asset == video_project_asset_unnormalized


def test_kili_convert_to_pixel_coords_image_rotated():
    """Test the conversion of coordinates from normalized to pixel values for rotated image files."""
    scaled_asset = convert_to_pixel_coords(
        image_asset_rotated, image_project, normalized_coordinates=False
    )
    assert scaled_asset == image_rotated_project_asset_unnormalized


@pytest.mark.parametrize(
    "json_interface, latest_label_annotations, expected_latest_label_result", test_cases
)
def test_video_object_detection_annotation_to_json_response(
    json_interface, latest_label_annotations, expected_latest_label_result
):
    """Test the conversion from annotations to jsonResponse."""
    with TemporaryDirectory() as tmp_dir:
        video_path = tmp_dir / Path("video1.mp4")
        video_path.write_bytes(b"fake video content")
        asset = {
            "id": "fake_asset_id",
            "resolution": {"width": 1920, "height": 1080},
            "content": str(video_path),
            "jsonContent": "",
        }

        converter = AnnotationsToJsonResponseConverter(
            json_interface=json_interface,
            project_input_type="VIDEO",
        )
        converter.patch_label_json_response(
            asset, latest_label_annotations, latest_label_annotations["annotations"]
        )
        del latest_label_annotations["annotations"]
        jobs = json_interface["jobs"].keys()
        job = next(iter(jobs))

        assert latest_label_annotations["labelType"] == expected_latest_label_result["labelType"]
        assert latest_label_annotations["author"] == expected_latest_label_result["author"]
        assert latest_label_annotations["jsonResponse"]["0"][job]["annotations"][0]["boundingPoly"][
            0
        ]["normalizedVertices"] == pytest.approx(
            expected_latest_label_result["jsonResponse"]["0"][job]["annotations"][0][
                "boundingPoly"
            ][0]["normalizedVertices"],
            rel=1e-2,
        )



from typing import Dict, List, Tuple

from kili_formats.tool.base import reverse_rotation_vertices
from kili_formats.types import JobCategory, JobTool


def convert_from_kili_to_yolo_format(
    job_id: str, label: Dict, category_ids: Dict[str, JobCategory]
) -> List[Tuple]:
    """Extract formatted annotations from labels and save the zip in the buckets."""
    if label is None or "jsonResponse" not in label:
        return []
    json_response = label["jsonResponse"]
    if not (job_id in json_response and "annotations" in json_response[job_id]):
        return []
    rotation_val = 0
    if "ROTATION_JOB" in json_response:
        rotation_val = json_response["ROTATION_JOB"]["rotation"]

    if not (job_id in json_response and "annotations" in json_response[job_id]):
        return []
    annotations = json_response[job_id]["annotations"]
    converted_annotations: List[Tuple] = []
    for annotation in annotations:
        category_idx: JobCategory = category_ids[
            get_category_full_name(job_id, annotation["categories"][0]["name"])
        ]
        if "boundingPoly" not in annotation:
            continue
        bounding_poly = annotation["boundingPoly"]
        if len(bounding_poly) < 1 or "normalizedVertices" not in bounding_poly[0]:
            continue
        normalized_vertices = bounding_poly[0]["normalizedVertices"]
        vertices_before_rotate = reverse_rotation_vertices(normalized_vertices, rotation_val)
        x_s: List[float] = [vertice["x"] for vertice in vertices_before_rotate]
        y_s: List[float] = [vertice["y"] for vertice in vertices_before_rotate]

        ## /!\ this part was only in the SDK to be tested
        if annotation["type"] == JobTool.RECTANGLE:
            x_min, y_min = min(x_s), min(y_s)
            x_max, y_max = max(x_s), max(y_s)
            bbox_center_x, bbox_center_y = (x_min + x_max) / 2, (y_min + y_max) / 2  # type: ignore
            bbox_width, bbox_height = x_max - x_min, y_max - y_min  # type: ignore
            converted_annotations.append(
                (category_idx.id, bbox_center_x, bbox_center_y, bbox_width, bbox_height)
            )

        elif annotation["type"] in {JobTool.POLYGON, JobTool.SEMANTIC}:
            # <class-index> <x1> <y1> <x2> <y2> ... <xn> <yn>
            # Each segmentation label must have a minimum of 3 xy points (polygon)
            points = [val for pair in zip(x_s, y_s) for val in pair]
            converted_annotations.append((category_idx.id, *points))

    return converted_annotations


def get_category_full_name(job_id, category_name):
    """Return a full name to identify uniquely a category."""
    return f"{job_id}__{category_name}"



from typing import Dict

from kili_formats.format.yolo import convert_from_kili_to_yolo_format
from kili_formats.types import JobCategory

from .fakes.yolo import (
    asset,
    asset_with_0_rotation,
    asset_with_90_rotation,
    asset_with_180_rotation,
    asset_with_270_rotation,
)

job_category_a: JobCategory = JobCategory(category_name="OBJECT_A", id=0, job_id="JOB_0")
job_category_b: JobCategory = JobCategory(category_name="OBJECT_B", id=1, job_id="JOB_0")
category_ids: Dict[str, JobCategory] = {
    "JOB_0__OBJECT_A": job_category_a,
    "JOB_0__OBJECT_B": job_category_b,
}


def test_convert_from_kili_to_yolo_format():
    converted_annotations = convert_from_kili_to_yolo_format(
        "JOB_0", asset["latestLabel"], category_ids
    )
    converted_annotations_with_rotation_0 = convert_from_kili_to_yolo_format(
        "JOB_0", asset_with_0_rotation["latestLabel"], category_ids
    )
    converted_annotations_with_rotation_90 = convert_from_kili_to_yolo_format(
        "JOB_0", asset_with_90_rotation["latestLabel"], category_ids
    )
    converted_annotations_with_rotation_180 = convert_from_kili_to_yolo_format(
        "JOB_0", asset_with_180_rotation["latestLabel"], category_ids
    )
    converted_annotations_with_rotation_270 = convert_from_kili_to_yolo_format(
        "JOB_0", asset_with_270_rotation["latestLabel"], category_ids
    )
    expected_annotations = [
        (
            0,
            0.501415026274802,
            0.5296278884310182,
            0.6727472455849373,
            0.5381320101586394,
        )
    ]
    expected_annotations2 = [
        (0, 0.20836785418392711, 0.28447691496573013, 0.2609776304888154, 0.3803570083603225)
    ]
    assert len(converted_annotations) == 1
    assert converted_annotations == expected_annotations
    assert len(converted_annotations_with_rotation_0) == 1
    assert converted_annotations_with_rotation_0 == expected_annotations2
    assert len(converted_annotations_with_rotation_90) == 1
    assert converted_annotations_with_rotation_90 == expected_annotations2
    assert len(converted_annotations_with_rotation_180) == 1
    assert converted_annotations_with_rotation_180 == expected_annotations2
    assert len(converted_annotations_with_rotation_270) == 1
    assert converted_annotations_with_rotation_270 == expected_annotations2



from typing import Dict


def kili_transcription_annotation_to_geojson_non_localised_feature(
    transcription_annotation: Dict, job_name: str
):
    return {
        "type": "Feature",
        "geometry": None,
        "properties": {
            "kili": {
                "text": transcription_annotation["text"],
                "job": job_name,
            },
        },
    }



"""
This module defines types and data structures used in Kili's annotation and job management system.
"""
from enum import Enum
from typing import (
    Any,
    Dict,
    List,
    Literal,
    NamedTuple,
    NewType,
    Optional,
    TypedDict,
    Union,
)

InputType = Literal[
    "GEOSPATIAL", "IMAGE", "LLM_INSTR_FOLLOWING", "LLM_RLHF", "LLM_STATIC", "PDF", "TEXT", "VIDEO"
]
MLTask = Literal["CLASSIFICATION", "NAMED_ENTITIES_RECOGNITION", "OBJECT_DETECTION"]

AnnotationId = NewType("AnnotationId", str)
AnnotationValueId = NewType("AnnotationValueId", str)
JobName = NewType("JobName", str)
KeyAnnotationId = NewType("KeyAnnotationId", str)
LabelId = NewType("LabelId", str)


class JobCategory(NamedTuple):
    """Contains information for a category."""

    category_name: str
    id: int
    job_id: str


class JobTool(str, Enum):
    """List of tools."""

    MARKER = "marker"
    POLYGON = "polygon"
    POLYLINE = "polyline"
    POSE = "pose"
    RANGE = "range"
    RECTANGLE = "rectangle"
    SEMANTIC = "semantic"
    VECTOR = "vector"


class Job(TypedDict):
    """Contains job settings."""

    content: Any
    instruction: str
    isChild: bool
    tools: List[JobTool]
    mlTask: MLTask
    models: Any  # example: {"interactive-segmentation": {"job": "SEMANTIC_JOB_MARKER"}},
    isVisible: bool
    required: int
    isNew: bool


class ProjectDict(TypedDict):
    """Dict that represents a Project."""

    description: str
    id: str
    inputType: InputType
    jsonInterface: Optional[Dict]
    organizationId: str
    title: str


class ChatItemRole(str, Enum):
    """Enumeration of the supported chat item role."""

    ASSISTANT = "ASSISTANT"
    USER = "USER"
    SYSTEM = "SYSTEM"


class ChatItem(TypedDict):
    """Dict that represents a ChatItem."""

    id: str
    content: str
    createdAt: Optional[str]
    externalId: str
    modelId: Optional[str]
    modelName: Optional[str]
    role: ChatItemRole


class ConversationLabel(TypedDict):
    """Dict that represents a ConversationLabel."""

    completion: Optional[Dict]
    conversation: Optional[Dict]
    round: Optional[Dict]


class Conversation(TypedDict):
    """Dict that represents a Conversation."""

    chatItems: List[ChatItem]
    externalId: Optional[str]
    label: Optional[ConversationLabel]
    labeler: Optional[str]
    metadata: Optional[dict]


class ExportLLMItem(TypedDict):
    """LLM asset chat part."""

    role: str
    content: str
    id: Optional[str]
    chat_id: Optional[str]
    model: Optional[str]


class JobLevel:
    """Job level."""

    ROUND = "round"
    CONVERSATION = "conversation"
    COMPLETION = "completion"


class Vertice(TypedDict):
    """Vertice."""

    x: float
    y: float


class ObjectDetectionAnnotationValue(TypedDict):
    """Object detection annotation value."""

    vertices: List[List[List[Vertice]]]


class ClassificationAnnotationValue(TypedDict):
    """Classification annotation value."""

    categories: List[str]


class ClassificationAnnotation(TypedDict):
    """Classification annotation."""

    # pylint: disable=unused-private-member
    __typename: Literal["ClassificationAnnotation"]
    id: AnnotationId
    labelId: LabelId
    job: JobName
    path: List[List[str]]
    annotationValue: ClassificationAnnotationValue


class RankingOrderValue(TypedDict):
    """Ranking order value."""

    rank: int
    elements: List[str]


class RankingAnnotationValue(TypedDict):
    """Ranking annotation value."""

    orders: List[RankingOrderValue]


class RankingAnnotation(TypedDict):
    """Ranking annotation."""

    # pylint: disable=unused-private-member
    __typename: Literal["RankingAnnotation"]
    id: AnnotationId
    labelId: LabelId
    job: JobName
    path: List[List[str]]
    annotationValue: RankingAnnotationValue


class TranscriptionAnnotationValue(TypedDict):
    """Transcription annotation value."""

    text: str


class TranscriptionAnnotation(TypedDict):
    """Transcription annotation."""

    # pylint: disable=unused-private-member
    __typename: Literal["TranscriptionAnnotation"]
    id: AnnotationId
    labelId: LabelId
    job: JobName
    path: List[List[str]]
    annotationValue: TranscriptionAnnotationValue


class Annotation(TypedDict):
    """Annotation."""

    id: AnnotationId
    labelId: LabelId
    job: JobName
    path: List[List[str]]


class ObjectDetectionAnnotation(TypedDict):
    """Object detection annotation."""

    # pylint: disable=unused-private-member
    __typename: Literal["ObjectDetectionAnnotation"]
    id: AnnotationId
    labelId: LabelId
    job: JobName
    path: List[List[str]]
    annotationValue: ObjectDetectionAnnotationValue
    name: Optional[str]
    mid: str
    category: str


class FrameInterval(TypedDict):
    """Frame interval."""

    start: int
    end: int


class VideoObjectDetectionKeyAnnotation(TypedDict):
    """Video object detection key annotation."""

    id: KeyAnnotationId
    frame: int
    annotationValue: ObjectDetectionAnnotationValue


class VideoClassificationKeyAnnotation(TypedDict):
    """Video classification key annotation."""

    id: KeyAnnotationId
    frame: int
    annotationValue: ClassificationAnnotationValue


class VideoTranscriptionKeyAnnotation(TypedDict):
    """Video transcription key annotation."""

    id: KeyAnnotationId
    frame: int
    annotationValue: TranscriptionAnnotationValue


class VideoObjectDetectionAnnotation(TypedDict):
    """Video object detection annotation."""

    # pylint: disable=unused-private-member
    __typename: Literal["VideoObjectDetectionAnnotation"]
    id: AnnotationId
    labelId: LabelId
    job: JobName
    path: List[List[str]]
    frames: List[FrameInterval]
    keyAnnotations: List[VideoObjectDetectionKeyAnnotation]
    name: Optional[str]
    mid: str
    category: str


class VideoClassificationAnnotation(TypedDict):
    """Video classification annotation."""

    # pylint: disable=unused-private-member
    __typename: Literal["VideoClassificationAnnotation"]
    id: AnnotationId
    labelId: LabelId
    job: JobName
    path: List[List[str]]
    frames: List[FrameInterval]
    keyAnnotations: List[VideoClassificationKeyAnnotation]


class VideoTranscriptionAnnotation(TypedDict):
    """Video transcription annotation."""

    # pylint: disable=unused-private-member
    __typename: Literal["VideoTranscriptionAnnotation"]
    id: AnnotationId
    labelId: LabelId
    job: JobName
    path: List[List[str]]
    frames: List[FrameInterval]
    keyAnnotations: List[VideoTranscriptionKeyAnnotation]


VideoAnnotation = Union[
    VideoObjectDetectionAnnotation,
    VideoClassificationAnnotation,
    VideoTranscriptionAnnotation,
]

ClassicAnnotation = Union[
    ClassificationAnnotation,
    ObjectDetectionAnnotation,
    RankingAnnotation,
    TranscriptionAnnotation,
]
